"""
These are designed to be unit-tests of the wrapper functionality. They do not test for
correctness of simulations,
but whether different parameter options work/don't work as intended.
"""
import pytest

import numpy as np

from py21cmfast import wrapper

REDSHIFT = 15
LOW_REDSHIFT = 8


@pytest.fixture(scope="module")
def user_params():
    # Do a small box, for testing
    return wrapper.UserParams(HII_DIM=35, DIM=70)


@pytest.fixture(scope="module")
def init_box(user_params, tmpdirec):
    return wrapper.initial_conditions(
        user_params=user_params, regenerate=True, direc=tmpdirec.strpath, random_seed=12
    )


@pytest.fixture(scope="module")
def perturb_field(init_box, tmpdirec):
    "A default perturb_field"
    return wrapper.perturb_field(
        redshift=REDSHIFT,
        regenerate=True,  # i.e. make sure we don't read it in.
        init_boxes=init_box,
        direc=tmpdirec.strpath,
    )


@pytest.fixture(scope="module")
def perturb_field_lowz(init_box, tmpdirec):
    "A default perturb_field"
    return wrapper.perturb_field(
        redshift=LOW_REDSHIFT,
        regenerate=True,  # i.e. make sure we don't read it in.
        init_boxes=init_box,
        direc=tmpdirec.strpath,
    )


@pytest.fixture(scope="module")
def ionize_box(perturb_field, tmpdirec):
    """A default ionize_box"""
    return wrapper.ionize_box(
        perturbed_field=perturb_field,
        regenerate=True,  # i.e. make sure we don't read it in.
        direc=tmpdirec.strpath,
    )


@pytest.fixture(scope="module")
def ionize_box_lowz(perturb_field_lowz, tmpdirec):
    """A default ionize_box at lower redshift."""
    return wrapper.ionize_box(
        perturbed_field=perturb_field_lowz,
        regenerate=True,  # i.e. make sure we don't read it in.
        direc=tmpdirec.strpath,
    )


@pytest.fixture(scope="module")
def spin_temp(perturb_field, tmpdirec):
    "A default perturb_field"
    return wrapper.spin_temperature(
        perturbed_field=perturb_field,
        regenerate=True,  # i.e. make sure we don't read it in.
        direc=tmpdirec.strpath,
    )


def test_perturb_field_no_ic(user_params, perturb_field):
    """Run a perturb field without passing an init box"""
    assert (
        len(perturb_field.density)
        == perturb_field.user_params.HII_DIM
        == user_params.HII_DIM
    )
    assert perturb_field.redshift == REDSHIFT
    assert not np.all(perturb_field.density == 0)


def test_ib_no_z(init_box):
    with pytest.raises(ValueError):
        wrapper.ionize_box(init_boxes=init_box)


def test_pf_unnamed_param():
    "Try using an un-named parameter"
    with pytest.raises(TypeError):
        wrapper.perturb_field(7)


def test_perturb_field_ic(user_params, perturb_field, tmpdirec):
    ic = wrapper.initial_conditions(
        user_params=user_params, regenerate=True, direc=tmpdirec.strpath, write=False
    )
    pf = wrapper.perturb_field(
        redshift=REDSHIFT,
        init_boxes=ic,
        regenerate=True,
        direc=tmpdirec.strpath,
        write=False,
    )

    assert len(pf.density) == len(ic.lowres_density)
    assert pf.cosmo_params == ic.cosmo_params
    assert pf.user_params == ic.user_params
    assert not np.all(pf.density == 0)

    assert pf.user_params == perturb_field.user_params
    assert pf.random_seed != perturb_field.random_seed
    assert pf.cosmo_params == perturb_field.cosmo_params

    # they shouldn't be the same, as they have different seeds
    assert pf != perturb_field

    # but they are the same in every other way
    assert pf._seedless_repr() == perturb_field._seedless_repr()


def test_cache_exists(user_params, perturb_field, tmpdirec):
    pf = wrapper.PerturbedField(
        redshift=perturb_field.redshift,
        cosmo_params=perturb_field.cosmo_params,
        user_params=user_params,
    )

    assert pf.exists(tmpdirec.strpath)

    pf.read(tmpdirec.strpath)
    assert np.all(pf.density == perturb_field.density)
    assert pf == perturb_field


def test_pf_new_seed(perturb_field, tmpdirec):
    pf = wrapper.perturb_field(
        redshift=perturb_field.redshift,
        user_params=perturb_field.user_params,
        direc=tmpdirec.strpath,
        random_seed=1,
        write=False,
    )

    assert not pf.exists(
        direc=tmpdirec.strpath
    )  # we didn't write it, and this has a different seed (presumably)
    assert pf.random_seed != perturb_field.random_seed

    assert not np.all(pf.density == perturb_field.density)


def test_ib_new_seed(ionize_box_lowz, perturb_field_lowz, tmpdirec):
    # this should fail because perturb_field has a seed set already, which isn't 1.
    with pytest.raises(ValueError):
        ib = wrapper.ionize_box(
            perturbed_field=perturb_field_lowz,
            direc=tmpdirec.strpath,
            random_seed=1,
            write=False,
        )

    ib = wrapper.ionize_box(
        cosmo_params=perturb_field_lowz.cosmo_params,
        redshift=perturb_field_lowz.redshift,
        user_params=perturb_field_lowz.user_params,
        direc=tmpdirec.strpath,
        random_seed=1,
        write=False,
    )

    # we didn't write it, and this has a different seed (presumably)
    assert not ib.exists(direc=tmpdirec.strpath)
    assert ib.random_seed != ionize_box_lowz.random_seed
    assert not np.all(ib.xH_box == ionize_box_lowz.xH_box)


def test_st_new_seed(spin_temp, perturb_field, tmpdirec):
    # this should fail because perturb_field has a seed set already, which isn't 1.
    with pytest.raises(ValueError):
        wrapper.spin_temperature(
            perturbed_field=perturb_field,
            direc=tmpdirec.strpath,
            random_seed=1,
            write=False,
        )

    st = wrapper.spin_temperature(
        cosmo_params=spin_temp.cosmo_params,
        user_params=spin_temp.user_params,
        astro_params=spin_temp.astro_params,
        flag_options=spin_temp.flag_options,
        redshift=spin_temp.redshift,
        direc=tmpdirec.strpath,
        random_seed=1,
        write=False,
    )

    assert not st.exists(
        direc=tmpdirec.strpath
    )  # we didn't write it, and this has a different seed (presumably)
    assert st.random_seed != spin_temp.random_seed
    assert not np.all(st.Ts_box == spin_temp.Ts_box)


def test_st_from_z(init_box, tmpdirec, spin_temp):
    pf = wrapper.perturb_field(
        redshift=12, init_boxes=init_box, write=False, regenerate=True
    )

    # This one has all the same parameters as the nominal spin_temp, but is evaluated with
    # perturb field exactly matching it, rather than interpolated
    st = wrapper.spin_temperature(
        perturbed_field=pf,
        astro_params=spin_temp.astro_params,
        flag_options=spin_temp.flag_options,
        direc=tmpdirec.strpath,
        redshift=spin_temp.redshift,
        write=False,
    )

    # TODO: This REALLY SHOULD NOT BE TRUE!!!!!
    assert st == spin_temp
    assert np.all(st.Ts_box == spin_temp.Ts_box)


def test_pf_regenerate(perturb_field, tmpdirec):
    pf = wrapper.perturb_field(
        redshift=perturb_field.redshift,
        user_params=perturb_field.user_params,
        direc=tmpdirec.strpath,
        regenerate=True,
    )

    assert not np.all(pf.density == perturb_field.density)
    assert pf.random_seed != perturb_field.random_seed


def test_ib_from_pf(perturb_field, tmpdirec):
    ib = wrapper.ionize_box(
        perturbed_field=perturb_field, direc=tmpdirec.strpath, write=False
    )
    assert ib.redshift == perturb_field.redshift
    assert ib.user_params == perturb_field.user_params
    assert ib.cosmo_params == perturb_field.cosmo_params


def test_ib_from_z(user_params, perturb_field, tmpdirec):
    ib = wrapper.ionize_box(
        redshift=perturb_field.redshift,
        user_params=user_params,
        direc=tmpdirec.strpath,
        write=False,
    )
    assert ib.redshift == perturb_field.redshift
    assert ib.user_params == perturb_field.user_params
    assert ib.cosmo_params == perturb_field.cosmo_params
    assert ib.cosmo_params is not perturb_field.cosmo_params


def test_ib_override_z(perturb_field, tmpdirec):
    with pytest.raises(ValueError):
        wrapper.ionize_box(
            redshift=perturb_field.redshift + 1,
            perturbed_field=perturb_field,
            direc=tmpdirec.strpath,
            write=False,
        )


def test_ib_override_z_heat_max(perturb_field, tmpdirec):
    # save previous z_heat_max
    zheatmax = wrapper.global_params.Z_HEAT_MAX

    wrapper.ionize_box(
        redshift=perturb_field.redshift,
        perturbed_field=perturb_field,
        direc=tmpdirec.strpath,
        write=False,
        z_heat_max=12.0,
    )

    assert wrapper.global_params.Z_HEAT_MAX == zheatmax


def test_ib_bad_st(init_box):
    with pytest.raises(ValueError):
        wrapper.ionize_box(redshift=REDSHIFT, spin_temp=init_box)


def test_bt(ionize_box, spin_temp, perturb_field):
    with pytest.raises(TypeError):  # have to specify param names
        wrapper.brightness_temperature(ionize_box, spin_temp, perturb_field)

    # this will fail because ionized_box was not created with spin temperature.
    with pytest.raises(ValueError):
        wrapper.brightness_temperature(
            ionized_box=ionize_box, perturbed_field=perturb_field, spin_temp=spin_temp
        )

    bt = wrapper.brightness_temperature(
        ionized_box=ionize_box, perturbed_field=perturb_field
    )

    assert bt.cosmo_params == perturb_field.cosmo_params
    assert bt.user_params == perturb_field.user_params
    assert bt.flag_options == ionize_box.flag_options
    assert bt.astro_params == ionize_box.astro_params


def test_coeval_against_direct(init_box, perturb_field, ionize_box):
    coeval = wrapper.run_coeval(perturb=perturb_field, init_box=init_box, write=False)

    assert coeval[0].init_struct == init_box
    assert coeval[0].perturb_struct == perturb_field
    assert coeval[0].ionization_struct == ionize_box


def test_lightcone(init_box, perturb_field):
    lc = wrapper.run_lightcone(
        init_box=init_box, perturb=perturb_field, max_redshift=10.0
    )

    assert lc.lightcone_redshifts[-1] >= 10.0
    assert np.isclose(lc.lightcone_redshifts[0], perturb_field.redshift, atol=1e-4)
    assert lc.cell_size == init_box.user_params.BOX_LEN / init_box.user_params.HII_DIM


def test_run_lf():
    muv, mhalo, lf = wrapper.compute_luminosity_function(redshifts=[7, 8, 9], nbins=100)
    assert np.all(lf[~np.isnan(lf)] > -30)
    assert lf.shape == (3, 100)

    # Check that memory is in-tact and a second run also works:
    muv, mhalo, lf2 = wrapper.compute_luminosity_function(
        redshifts=[7, 8, 9], nbins=100
    )
    assert lf2.shape == (3, 100)
    assert np.allclose(lf2[~np.isnan(lf2)], lf[~np.isnan(lf)])



"""Build the C code with CFFI."""
import os

from cffi import FFI

ffi = FFI()
LOCATION = os.path.dirname(os.path.abspath(__file__))
CLOC = os.path.join(LOCATION, "src", "py21cmfast", "src")
include_dirs = [CLOC]

# =================================================================
# Set compilation arguments dependent on environment... a bit buggy
# =================================================================
if "DEBUG" in os.environ:
    extra_compile_args = ["-fopenmp", "-w", "-g", "-O0"]
else:
    extra_compile_args = ["-fopenmp", "-Ofast", "-w"]

# Set the C-code logging level.
# If DEBUG is set, we default to the highest level, but if not,
# we set it to the level just above no logging at all.
log_level = os.environ.get("LOG_LEVEL", 3 if "DEBUG" in os.environ else 1)
available_levels = [
    "NONE",
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
    "SUPER_DEBUG",
    "ULTRA_DEBUG",
]


if isinstance(log_level, str) and log_level.upper() in available_levels:
    log_level = available_levels.index(log_level.upper())

try:
    log_level = int(log_level)
except ValueError:
    # note: for py35 support, can't use f strings.
    raise ValueError(
        "LOG_LEVEL must be specified as a positive integer, or one "
        "of {}".format(available_levels)
    )

library_dirs = []
for k, v in os.environ.items():
    if "inc" in k.lower():
        include_dirs += [v]
    elif "lib" in k.lower():
        library_dirs += [v]

# =================================================================

# This is the overall C code.
ffi.set_source(
    "py21cmfast.c_21cmfast",  # Name/Location of shared library module
    """
    #define LOG_LEVEL {log_level}

    #include "GenerateICs.c"
    """.format(
        log_level=log_level
    ),
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=["m", "gsl", "gslcblas", "fftw3f_omp", "fftw3f"],
    extra_compile_args=extra_compile_args,
    extra_link_args=["-fopenmp"],
)

# This is the Header file
with open(os.path.join(CLOC, "21cmFAST.h")) as f:
    ffi.cdef(f.read())

with open(os.path.join(CLOC, "Globals.h")) as f:
    ffi.cdef(f.read())

if __name__ == "__main__":
    ffi.compile()



"""
A set of large-scale tests which test code updates against previously-run "golden"
results.

The idea here is that any new updates (except for major versions) should be non-breaking;
firstly, they should not break the API, so that the tests should run without crashing without
being changed.
Secondly, the actual results of running the basic functions should remain the same for
the same input code, except for potential bug-fixes. In these cases, these tests should
pick these changes up. The test data should then be changed to reflect the new gold
standard, and if applicable, a new test should be written that reflects the previous
broken code.
Thirdly, it enforces that new features, where possible, are added in such a way as to
keep the default behaviour constant. That is, the tests here should *not* run the added
feature, and therefore should continue to produce the same test results regardless of
the new feature added. The new feature should be accompanied by its own tests, whether
in this or another test module. If a new feature *must* be included by default, then
it must be implemented in a new major version of the code, at which point the test data
is able to be updated.

Comparison tests here are meant to be as small as possible while attempting to form
a reasonable test: they should be of reduced data such as power spectra or global xHI
measurements, and they should be generated with small simulations.
"""
import pytest

import h5py
import numpy as np

from py21cmfast import global_params

from . import produce_integration_test_data as prd


@pytest.mark.parametrize("redshift,kwargs", prd.OPTIONS)
def test_power_spectra_coeval(redshift, kwargs):
    print("Options used for the test: ", kwargs)

    # First get pre-made data
    with h5py.File(prd.get_filename(redshift, **kwargs), "r") as f:
        power = f["power_coeval"][...]

    with global_params.use(zprime_step_factor=prd.DEFAULT_ZPRIME_STEP_FACTOR):
        # Note that if zprime_step_factor is set in kwargs, it will over-ride this.
        k, p, bt = prd.produce_coeval_power_spectra(redshift, **kwargs)

    assert np.allclose(power, p, atol=1e-5, rtol=1e-3)


@pytest.mark.parametrize("redshift,kwargs", prd.OPTIONS)
def test_power_spectra_lightcone(redshift, kwargs):
    print("Options used for the test: ", kwargs)

    # First get pre-made data
    with h5py.File(prd.get_filename(redshift, **kwargs), "r") as f:
        power = f["power_lc"][...]
        xHI = f["xHI"][...]
        Tb = f["Tb"][...]

    with global_params.use(zprime_step_factor=prd.DEFAULT_ZPRIME_STEP_FACTOR):
        # Note that if zprime_step_factor is set in kwargs, it will over-ride this.
        k, p, lc = prd.produce_lc_power_spectra(redshift, **kwargs)

    assert np.allclose(power, p, atol=1e-5, rtol=5e-3)
    assert np.allclose(xHI, lc.global_xHI, atol=1e-5, rtol=1e-3)
    assert np.allclose(Tb, lc.global_brightness_temp, atol=1e-5, rtol=1e-3)



"""Simple plotting functions for 21cmFAST objects."""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors

from . import outputs
from .wrapper import Coeval
from .wrapper import LightCone

eor_colour = colors.LinearSegmentedColormap.from_list(
    "EoR",
    [
        (0, "white"),
        (0.21, "yellow"),
        (0.42, "orange"),
        (0.63, "red"),
        (0.86, "black"),
        (0.9, "blue"),
        (1, "cyan"),
    ],
)
plt.register_cmap(cmap=eor_colour)


def _imshow_slice(
    cube,
    slice_axis=-1,
    slice_index=0,
    fig=None,
    ax=None,
    fig_kw=None,
    cbar=True,
    cbar_horizontal=False,
    rotate=False,
    cmap="EoR",
    **imshow_kw,
):
    """
    Plot a slice of some kind of cube.

    Parameters
    ----------
    cube : nd-array
        A 3D array of some quantity.
    slice_axis : int, optional
        The axis over which to take a slice, in order to plot.
    slice_index :
        The index of the slice.
    fig : Figure object
        An optional matplotlib figure object on which to plot
    ax : Axis object
        The matplotlib axis object on which to plot (created by default).
    fig_kw :
        Optional arguments passed to the figure construction.
    cbar : bool
        Whether to plot the colorbar
    cbar_horizontal : bool
        Whether the colorbar should be horizontal underneath the plot.
    rotate : bool
        Whether to rotate the plot vertically.
    imshow_kw :
        Optional keywords to pass to :func:`maplotlib.imshow`.

    Returns
    -------
    fig, ax :
        The figure and axis objects from matplotlib.
    """
    # If no axis is passed, create a new one
    # This allows the user to add this plot into an existing grid, or alter it afterwards.
    if fig_kw is None:
        fig_kw = {}
    if ax is None and fig is None:
        fig, ax = plt.subplots(1, 1, **fig_kw)
    elif ax is None:
        ax = plt.gca()
    elif fig is None:
        fig = plt.gcf()

    plt.sca(ax)

    if slice_index >= cube.shape[slice_axis]:
        raise IndexError(
            "slice_index is too large for that axis (slice_index=%s >= %s"
            % (slice_index, cube.shape[slice_axis])
        )

    slc = np.take(cube, slice_index, axis=slice_axis)
    if not rotate:
        slc = slc.T

    if cmap == "EoR":
        imshow_kw["vmin"] = -150
        imshow_kw["vmax"] = 30

    plt.imshow(slc, origin="lower", cmap=cmap, **imshow_kw)

    if cbar:
        cb = plt.colorbar(
            orientation="horizontal" if cbar_horizontal else "vertical", aspect=40
        )
        cb.outline.set_edgecolor(None)

    return fig, ax


def coeval_sliceplot(
    struct: [outputs._OutputStruct, Coeval], kind: [str, None] = None, **kwargs
):
    """
    Show a slice of a given coeval box.

    Parameters
    ----------
    struct : :class:`~outputs._OutputStruct` or :class:`~wrapper.Coeval` instance
        The output of a function such as `ionize_box` (a class containing several quantities), or
        `run_coeval`.
    kind : str
        The quantity within the structure to be shown.


    Returns
    -------
    fig, ax :
        figure and axis objects from matplotlib

    Other Parameters
    ----------------
    All other parameters are passed directly to :func:`_imshow_slice`. These include `slice_axis`
    and `slice_index`,
    which choose the actual slice to plot, optional `fig` and `ax` keywords which enable
    over-plotting previous figures,
    and the `imshow_kw` argument, which allows arbitrary styling of the plot.
    """
    if kind is None and isinstance(struct, outputs._OutputStruct):
        kind = struct.fieldnames[0]
    elif kind is None and isinstance(struct, Coeval):
        kind = "brightness_temp"

    try:
        cube = getattr(struct, kind)
    except AttributeError:
        raise AttributeError(
            "The given OutputStruct does not have the quantity {kind}".format(kind=kind)
        )

    if kind != "brightness_temp" and "cmap" not in kwargs:
        kwargs["cmap"] = "viridis"

    fig, ax = _imshow_slice(cube, extent=(0, struct.user_params.BOX_LEN) * 2, **kwargs)

    slice_axis = kwargs.get("slice_axis", -1)

    # Determine which axes are being plotted.
    if slice_axis in (2, -1):
        xax = "x"
        yax = "y"
    elif slice_axis == 1:
        xax = "x"
        yax = "z"
    elif slice_axis == 0:
        xax = "y"
        yax = "z"
    else:
        raise ValueError("slice_axis should be between -1 and 2")

    # Now put on the decorations.
    ax.set_xlabel("{xax}-axis [Mpc]".format(xax=xax))
    ax.set_ylabel("{yax}-axis [Mpc]".format(yax=yax))

    return fig, ax


def lightcone_sliceplot(
    lightcone: LightCone,
    kind: str = "brightness_temp",
    lightcone2: LightCone = None,
    vertical: bool = False,
    **kwargs,
):
    """Create a 2D plot of a slice through a lightcone.

    Parameters
    ----------
    lightcone : :class:`~py21cmfast.wrapper.Lightcone`
        The lightcone object to plot
    kind : str, optional
        The attribute of the lightcone to plot. Must be an array.
    lightcone2 : str, optional
        If provided, plot the _difference_ of the selected attribute between the two
        lightcones.
    vertical : bool, optional
        Whether to plot the redshift in the vertical direction.
    kwargs :
        Passed through to ``imshow()``.

    Returns
    -------
    fig :
        The matplotlib Figure object
    ax :
        The matplotlib Axis object onto which the plot was drawn.
    """
    slice_axis = kwargs.pop("slice_axis", 0)

    if slice_axis == 0:
        xlabel = "Redshift Axis [Mpc]"
        ylabel = "y-axis [Mpc]"

        if vertical:
            extent = (
                0,
                lightcone.user_params.BOX_LEN,
                0,
                lightcone.lightcone_coords[-1],
            )
            xlabel, ylabel = ylabel, xlabel
        else:
            extent = (
                0,
                lightcone.lightcone_coords[-1],
                0,
                lightcone.user_params.BOX_LEN,
            )

    else:
        extent = (0, lightcone.user_params.BOX_LEN) * 2

        if slice_axis == 1:
            xlabel = "x-axis [Mpc]"
            ylabel = "Redshift Axis [Mpc]"

        elif slice_axis in (2, -1):
            xlabel = "x-axis [Mpc]"
            ylabel = "y-axis [Mpc]"
        else:
            raise ValueError("slice_axis must be between -1 and 2")

    if lightcone2 is None:
        fig, ax = _imshow_slice(
            getattr(lightcone, kind),
            extent=extent,
            slice_axis=slice_axis,
            rotate=not vertical,
            cbar_horizontal=not vertical,
            **kwargs,
        )
    else:
        d = getattr(lightcone, kind) - getattr(lightcone2, kind)
        fig, ax = _imshow_slice(
            d,
            extent=extent,
            slice_axis=slice_axis,
            rotate=not vertical,
            cbar_horizontal=not vertical,
            cmap=kwargs.pop("cmap", "bwr"),
            vmin=-np.abs(d.max()),
            vmax=np.abs(d.max()),
            **kwargs,
        )
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    # TODO: use twinx to put a redshift axis on it.
    return fig, ax
